@@ -17,6 +17,7 @@
 import contextlib
 import datetime
 import errno
+import functools
 import hashlib
 import logging
 import os
@@ -27,8 +28,6 @@ import tempfile
 import threading
 import weakref
 
-from concurrent import futures
-
 import fasteners
 from oslo_utils import encodeutils
 from oslo_utils import fileutils
@@ -48,6 +47,8 @@ class _Barrier(object):
     def __init__(self):
         self.cond = threading.Condition()
         self.owner = None
+        self.shared = False
+        self.ref = 0
 
 
 @contextlib.contextmanager
@@ -55,9 +56,9 @@ def _translate_failures():
     try:
         yield
     except (EnvironmentError, voluptuous.Invalid) as e:
-        coordination.raise_with_cause(coordination.ToozError,
-                                      encodeutils.exception_to_unicode(e),
-                                      cause=e)
+        utils.raise_with_cause(tooz.ToozError,
+                               encodeutils.exception_to_unicode(e),
+                               cause=e)
 
 
 def _convert_from_old_format(data):
@@ -109,11 +110,12 @@ class FileLock(locking.Lock):
         self._lock = fasteners.InterProcessLock(path)
         self._barrier = barrier
         self._member_id = member_id
+        self.ref = 0
 
     def is_still_owner(self):
         return self.acquired
 
-    def acquire(self, blocking=True):
+    def acquire(self, blocking=True, shared=False):
         blocking, timeout = utils.convert_blocking(blocking)
         watch = timeutils.StopWatch(duration=timeout)
         watch.start()
@@ -121,11 +123,16 @@ class FileLock(locking.Lock):
         # Make the shared barrier ours first.
         with self._barrier.cond:
             while self._barrier.owner is not None:
+                if (shared and self._barrier.shared):
+                    break
                 if not blocking or watch.expired():
                     return False
                 self._barrier.cond.wait(watch.leftover(return_none=True))
             self._barrier.owner = (threading.current_thread().ident,
                                    os.getpid(), self._member_id)
+            self._barrier.shared = shared
+            self._barrier.ref += 1
+            self.ref += 1
 
         # Ok at this point we are now working in a thread safe manner,
         # and now we can try to get the actual lock...
@@ -144,6 +151,8 @@ class FileLock(locking.Lock):
                 # Release the barrier to let someone else have a go at it...
                 with self._barrier.cond:
                     self._barrier.owner = None
+                    self._barrier.ref = 0
+                    self._barrier.shared = False
                     self._barrier.cond.notify_all()
 
         self.acquired = gotten
@@ -153,9 +162,14 @@ class FileLock(locking.Lock):
         if not self.acquired:
             return False
         with self._barrier.cond:
-            self.acquired = False
-            self._barrier.owner = None
-            self._barrier.cond.notify_all()
+            self._barrier.ref -= 1
+            self.ref -= 1
+            if not self.ref:
+                self.acquired = False
+            if not self._barrier.ref:
+                self._barrier.owner = None
+                self._lock.release()
+                self._barrier.cond.notify_all()
         return True
 
     def __del__(self):
@@ -163,7 +177,8 @@ class FileLock(locking.Lock):
             LOG.warning("Unreleased lock %s garbage collected", self.name)
 
 
-class FileDriver(coordination.CoordinationDriverCachedRunWatchers):
+class FileDriver(coordination.CoordinationDriverCachedRunWatchers,
+                 coordination.CoordinationDriverWithExecutor):
     """A file based driver.
 
     This driver uses files and directories (and associated file locks) to
@@ -205,9 +220,8 @@ class FileDriver(coordination.CoordinationDriverCachedRunWatchers):
 
     def __init__(self, member_id, parsed_url, options):
         """Initialize the file driver."""
-        super(FileDriver, self).__init__(member_id)
+        super(FileDriver, self).__init__(member_id, parsed_url, options)
         self._dir = self._normalize_path(parsed_url.path)
-        self._executor = utils.ProxyExecutor.build("File", options)
         self._group_dir = os.path.join(self._dir, 'groups')
         self._tmpdir = os.path.join(self._dir, 'tmp')
         self._driver_lock_path = os.path.join(self._dir, '.driver_lock')
@@ -217,6 +231,7 @@ class FileDriver(coordination.CoordinationDriverCachedRunWatchers):
         self._reserved_paths = list(self._reserved_dirs)
         self._reserved_paths.append(self._driver_lock_path)
         self._safe_member_id = self._make_filesystem_safe(member_id)
+        self._timeout = int(self._options.get('timeout', 10))
 
     @staticmethod
     def _normalize_path(path):
@@ -250,15 +265,12 @@ class FileDriver(coordination.CoordinationDriverCachedRunWatchers):
         return hashlib.new(cls.HASH_ROUTINE, item).hexdigest()
 
     def _start(self):
+        super(FileDriver, self)._start()
         for a_dir in self._reserved_dirs:
             try:
                 fileutils.ensure_tree(a_dir)
             except OSError as e:
                 raise coordination.ToozConnectionError(e)
-        self._executor.start()
-
-    def _stop(self):
-        self._executor.stop()
 
     def _update_group_metadata(self, path, group_id):
         details = {
@@ -390,7 +402,15 @@ class FileDriver(coordination.CoordinationDriverCachedRunWatchers):
                         continue
                     entry_path = os.path.join(group_dir, entry)
                     try:
-                        member_id = self._read_member_id(entry_path)
+                        m_time = datetime.datetime.fromtimestamp(
+                            os.stat(entry_path).st_mtime)
+                        current_time = datetime.datetime.now()
+                        delta_time = timeutils.delta_seconds(m_time,
+                                                             current_time)
+                        if delta_time >= 0 and delta_time <= self._timeout:
+                            member_id = self._read_member_id(entry_path)
+                        else:
+                            continue
                     except EnvironmentError as e:
                         if e.errno != errno.ENOENT:
                             raise
@@ -445,7 +465,7 @@ class FileDriver(coordination.CoordinationDriverCachedRunWatchers):
                 if len(entries) > 1:
                     raise coordination.GroupNotEmpty(group_id)
                 elif len(entries) == 1 and entries != ['.metadata']:
-                    raise coordination.ToozError(
+                    raise tooz.ToozError(
                         "Unexpected path '%s' found in"
                         " group directory '%s' (expected to only find"
                         " a '.metadata' path)" % (entries[0], group_dir))
@@ -482,6 +502,23 @@ class FileDriver(coordination.CoordinationDriverCachedRunWatchers):
         fut = self._executor.submit(_do_get_groups)
         return FileFutureResult(fut)
 
+    def heartbeat(self):
+        for group_id in self._joined_groups:
+            safe_group_id = self._make_filesystem_safe(group_id)
+            group_dir = os.path.join(self._group_dir, safe_group_id)
+            member_path = os.path.join(group_dir, "%s.raw" %
+                                       self._safe_member_id)
+
+            @_lock_me(self._driver_lock)
+            def _do_heartbeat():
+                try:
+                    os.utime(member_path, None)
+                except EnvironmentError as err:
+                    if err.errno != errno.ENOENT:
+                        raise
+            _do_heartbeat()
+        return self._timeout
+
     @staticmethod
     def watch_elected_as_leader(group_id, callback):
         raise tooz.NotImplemented
@@ -491,23 +528,5 @@ class FileDriver(coordination.CoordinationDriverCachedRunWatchers):
         raise tooz.NotImplemented
 
 
-class FileFutureResult(coordination.CoordAsyncResult):
-    """File asynchronous result that references a future."""
-
-    def __init__(self, fut):
-        self._fut = fut
-
-    def get(self, timeout=None):
-        try:
-            # Late translate the common failures since the file driver
-            # may throw things that we can not catch in the callbacks where
-            # it is used.
-            with _translate_failures():
-                return self._fut.result(timeout=timeout)
-        except futures.TimeoutError as e:
-            coordination.raise_with_cause(coordination.OperationTimedOut,
-                                          encodeutils.exception_to_unicode(e),
-                                          cause=e)
-
-    def done(self):
-        return self._fut.done()
+FileFutureResult = functools.partial(coordination.CoordinatorResult,
+                                     failure_translator=_translate_failures)
