@@ -14,11 +14,12 @@
 # License for the specific language governing permissions and limitations
 # under the License.
 
+import contextlib
 import errno
+import functools
 import logging
 import socket
 
-from concurrent import futures
 from oslo_utils import encodeutils
 from pymemcache import client as pymemcache_client
 import six
@@ -33,35 +34,41 @@ from tooz import utils
 LOG = logging.getLogger(__name__)
 
 
-def _translate_failures(func):
+@contextlib.contextmanager
+def _failure_translator():
     """Translates common pymemcache exceptions into tooz exceptions.
 
     https://github.com/pinterest/pymemcache/blob/d995/pymemcache/client.py#L202
     """
+    try:
+        yield
+    except pymemcache_client.MemcacheUnexpectedCloseError as e:
+        utils.raise_with_cause(coordination.ToozConnectionError,
+                               encodeutils.exception_to_unicode(e),
+                               cause=e)
+    except (socket.timeout, socket.error,
+            socket.gaierror, socket.herror) as e:
+        # TODO(harlowja): get upstream pymemcache to produce a better
+        # exception for these, using socket (vs. a memcache specific
+        # error) seems sorta not right and/or the best approach...
+        msg = encodeutils.exception_to_unicode(e)
+        if e.errno is not None:
+            msg += " (with errno %s [%s])" % (errno.errorcode[e.errno],
+                                              e.errno)
+        utils.raise_with_cause(coordination.ToozConnectionError,
+                               msg, cause=e)
+    except pymemcache_client.MemcacheError as e:
+        utils.raise_with_cause(tooz.ToozError,
+                               encodeutils.exception_to_unicode(e),
+                               cause=e)
+
+
+def _translate_failures(func):
 
     @six.wraps(func)
     def wrapper(*args, **kwargs):
-        try:
+        with _failure_translator():
             return func(*args, **kwargs)
-        except pymemcache_client.MemcacheUnexpectedCloseError as e:
-            utils.raise_with_cause(coordination.ToozConnectionError,
-                                   encodeutils.exception_to_unicode(e),
-                                   cause=e)
-        except (socket.timeout, socket.error,
-                socket.gaierror, socket.herror) as e:
-            # TODO(harlowja): get upstream pymemcache to produce a better
-            # exception for these, using socket (vs. a memcache specific
-            # error) seems sorta not right and/or the best approach...
-            msg = encodeutils.exception_to_unicode(e)
-            if e.errno is not None:
-                msg += " (with errno %s [%s])" % (errno.errorcode[e.errno],
-                                                  e.errno)
-            utils.raise_with_cause(coordination.ToozConnectionError,
-                                   msg, cause=e)
-        except pymemcache_client.MemcacheError as e:
-            utils.raise_with_cause(tooz.ToozError,
-                                   encodeutils.exception_to_unicode(e),
-                                   cause=e)
 
     return wrapper
 
@@ -83,7 +90,9 @@ class MemcachedLock(locking.Lock):
                 return False
             return owner == self.coord._member_id
 
-    def acquire(self, blocking=True):
+    def acquire(self, blocking=True, shared=False):
+        if shared:
+            raise tooz.NotImplemented
 
         @_retry.retry(stop_max_delay=blocking)
         @_translate_failures
@@ -156,10 +165,12 @@ class MemcachedLock(locking.Lock):
             poked = self.coord.client.touch(self.name,
                                             expire=self.timeout,
                                             noreply=False)
-            if not poked:
-                LOG.warning("Unable to heartbeat by updating key '%s' with "
-                            "extended expiry of %s seconds", self.name,
-                            self.timeout)
+            if poked:
+                return True
+            LOG.warning("Unable to heartbeat by updating key '%s' with "
+                        "extended expiry of %s seconds", self.name,
+                        self.timeout)
+        return False
 
     @_translate_failures
     def get_owner(self):
@@ -170,7 +181,8 @@ class MemcachedLock(locking.Lock):
         return self in self.coord._acquired_locks
 
 
-class MemcachedDriver(coordination.CoordinationDriverCachedRunWatchers):
+class MemcachedDriver(coordination.CoordinationDriverCachedRunWatchers,
+                      coordination.CoordinationDriverWithExecutor):
     """A `memcached`_ based driver.
 
     This driver users `memcached`_ concepts to provide the coordination driver
@@ -217,21 +229,18 @@ class MemcachedDriver(coordination.CoordinationDriverCachedRunWatchers):
     STILL_ALIVE = b"It's alive!"
 
     def __init__(self, member_id, parsed_url, options):
-        super(MemcachedDriver, self).__init__(member_id)
-        options = utils.collapse(options)
-        self._options = options
-        self._executor = utils.ProxyExecutor.build("Memcached", options)
+        super(MemcachedDriver, self).__init__(member_id, parsed_url, options)
         self.host = (parsed_url.hostname or "localhost",
                      parsed_url.port or 11211)
-        default_timeout = options.get('timeout', self.DEFAULT_TIMEOUT)
+        default_timeout = self._options.get('timeout', self.DEFAULT_TIMEOUT)
         self.timeout = int(default_timeout)
-        self.membership_timeout = int(options.get(
+        self.membership_timeout = int(self._options.get(
             'membership_timeout', default_timeout))
-        self.lock_timeout = int(options.get(
+        self.lock_timeout = int(self._options.get(
             'lock_timeout', default_timeout))
-        self.leader_timeout = int(options.get(
+        self.leader_timeout = int(self._options.get(
             'leader_timeout', default_timeout))
-        max_pool_size = options.get('max_pool_size', None)
+        max_pool_size = self._options.get('max_pool_size', None)
         if max_pool_size is not None:
             self.max_pool_size = int(max_pool_size)
         else:
@@ -255,6 +264,7 @@ class MemcachedDriver(coordination.CoordinationDriverCachedRunWatchers):
 
     @_translate_failures
     def _start(self):
+        super(MemcachedDriver, self)._start()
         self.client = pymemcache_client.PooledClient(
             self.host,
             serializer=self._msgpack_serializer,
@@ -265,14 +275,13 @@ class MemcachedDriver(coordination.CoordinationDriverCachedRunWatchers):
         # Run heartbeat here because pymemcache use a lazy connection
         # method and only connect once you do an operation.
         self.heartbeat()
-        self._executor.start()
 
     @_translate_failures
     def _stop(self):
+        super(MemcachedDriver, self)._stop()
         for lock in list(self._acquired_locks):
             lock.release()
         self.client.delete(self._encode_member_id(self._member_id))
-        self._executor.stop()
         self.client.close()
 
     def _encode_group_id(self, group_id):
@@ -502,19 +511,6 @@ class MemcachedDriver(coordination.CoordinationDriverCachedRunWatchers):
         return result
 
 
-class MemcachedFutureResult(coordination.CoordAsyncResult):
-    """Memcached asynchronous result that references a future."""
-    def __init__(self, fut):
-        self._fut = fut
-
-    def get(self, timeout=10):
-        try:
-            return self._fut.result(timeout=timeout)
-        except futures.TimeoutError as e:
-            utils.raise_with_cause(
-                coordination.OperationTimedOut,
-                encodeutils.exception_to_unicode(e),
-                cause=e)
-
-    def done(self):
-        return self._fut.done()
+MemcachedFutureResult = functools.partial(
+    coordination.CoordinatorResult,
+    failure_translator=_failure_translator)
