@@ -124,7 +124,7 @@ class TestAPI(tests.TestWithCoordinator):
         delete = self._coord.delete_group(self.group_id)
         self.assertRaises(tooz.coordination.GroupNotEmpty,
                           delete.get)
-        self._coord.leave_group(self.group_id)
+        self._coord.leave_group(self.group_id).get()
         self._coord.delete_group(self.group_id).get()
 
     def test_join_group(self):
@@ -221,7 +221,7 @@ class TestAPI(tests.TestWithCoordinator):
             'type': 'warrior',
             'abilities': ['fight', 'flight', 'double-hit-damage'],
         }
-        self._coord.join_group(self.group_id, caps)
+        self._coord.join_group(self.group_id, caps).get()
         capa = self._coord.get_member_capabilities(self.group_id,
                                                    self.member_id).get()
         self.assertEqual(capa, caps)
@@ -244,7 +244,7 @@ class TestAPI(tests.TestWithCoordinator):
 
     def test_get_member_info(self):
         self._coord.create_group(self.group_id).get()
-        self._coord.join_group(self.group_id, b"test_capabilities")
+        self._coord.join_group(self.group_id, b"test_capabilities").get()
 
         member_info = self._coord.get_member_info(self.group_id,
                                                   self.member_id).get()
@@ -259,7 +259,7 @@ class TestAPI(tests.TestWithCoordinator):
         member_info = {'capabilities': 'caps',
                        'created_at': '0',
                        'updated_at': '0'}
-        self._coord.join_group(self.group_id, caps)
+        self._coord.join_group(self.group_id, caps).get()
         member_info = self._coord.get_member_info(self.group_id,
                                                   self.member_id).get()
         self.assertEqual(member_info['capabilities'], caps)
@@ -712,6 +712,68 @@ class TestAPI(tests.TestWithCoordinator):
         with lock:
             pass
 
+    def test_heartbeat_lock_not_acquired(self):
+        lock = self._coord.get_lock(tests.get_random_uuid())
+        # Not all locks need heartbeat
+        if hasattr(lock, "heartbeat"):
+            self.assertFalse(lock.heartbeat())
+
+    def test_get_shared_lock(self):
+        lock = self._coord.get_lock(tests.get_random_uuid())
+        self.assertTrue(lock.acquire(shared=True))
+        self.assertTrue(lock.release())
+        with lock(shared=True):
+            pass
+
+    def test_get_shared_lock_locking_same_lock_twice(self):
+        lock = self._coord.get_lock(tests.get_random_uuid())
+        self.assertTrue(lock.acquire(shared=True))
+        self.assertTrue(lock.acquire(shared=True))
+        self.assertTrue(lock.release())
+        self.assertTrue(lock.release())
+        self.assertFalse(lock.release())
+        with lock(shared=True):
+            pass
+
+    def test_get_shared_lock_locking_two_lock(self):
+        name = tests.get_random_uuid()
+        lock1 = self._coord.get_lock(name)
+        coord = tooz.coordination.get_coordinator(
+            self.url, tests.get_random_uuid())
+        coord.start()
+        lock2 = coord.get_lock(name)
+
+        self.assertTrue(lock1.acquire(shared=True))
+        self.assertTrue(lock2.acquire(shared=True))
+        self.assertTrue(lock1.release())
+        self.assertTrue(lock2.release())
+
+    def test_get_lock_locking_shared_and_exclusive(self):
+        name = tests.get_random_uuid()
+        lock1 = self._coord.get_lock(name)
+        coord = tooz.coordination.get_coordinator(
+            self.url, tests.get_random_uuid())
+        coord.start()
+        lock2 = coord.get_lock(name)
+
+        self.assertTrue(lock1.acquire(shared=True))
+        self.assertFalse(lock2.acquire(blocking=False))
+        self.assertTrue(lock1.release())
+        self.assertFalse(lock2.release())
+
+    def test_get_lock_locking_exclusive_and_shared(self):
+        name = tests.get_random_uuid()
+        lock1 = self._coord.get_lock(name)
+        coord = tooz.coordination.get_coordinator(
+            self.url, tests.get_random_uuid())
+        coord.start()
+        lock2 = coord.get_lock(name)
+
+        self.assertTrue(lock1.acquire())
+        self.assertFalse(lock2.acquire(shared=True, blocking=False))
+        self.assertTrue(lock1.release())
+        self.assertFalse(lock2.release())
+
     def test_get_lock_concurrency_locking_same_lock(self):
         lock = self._coord.get_lock(tests.get_random_uuid())
 
@@ -743,10 +805,38 @@ class TestAPI(tests.TestWithCoordinator):
                              tests.get_random_uuid())
                 self.assertFalse(f.result())
 
+    def _do_test_get_lock_serial_locking_two_lock(self, executor,
+                                                  use_same_coord):
+        name = tests.get_random_uuid()
+        lock1 = self._coord.get_lock(name)
+        lock1.acquire()
+        lock1.release()
+        with executor(max_workers=1) as e:
+            coord = self._coord if use_same_coord else None
+            f = e.submit(try_to_lock_job, name, coord, self.url,
+                         tests.get_random_uuid())
+            self.assertTrue(f.result())
+
     def test_get_lock_concurrency_locking_two_lock_process(self):
+        # NOTE(jd) Using gRPC and forking is not supported so this test might
+        # very likely hang forever or crash. See
+        # https://github.com/grpc/grpc/issues/10140#issuecomment-297548714 for
+        # more info.
+        if self.url.startswith("etcd3://"):
+            self.skipTest("Unable to use etcd3 with fork()")
         self._do_test_get_lock_concurrency_locking_two_lock(
             futures.ProcessPoolExecutor, False)
 
+    def test_get_lock_serial_locking_two_lock_process(self):
+        # NOTE(jd) Using gRPC and forking is not supported so this test might
+        # very likely hang forever or crash. See
+        # https://github.com/grpc/grpc/issues/10140#issuecomment-297548714 for
+        # more info.
+        if self.url.startswith("etcd3://"):
+            self.skipTest("Unable to use etcd3 with fork()")
+        self._do_test_get_lock_serial_locking_two_lock(
+            futures.ProcessPoolExecutor, False)
+
     def test_get_lock_concurrency_locking_two_lock_thread1(self):
         self._do_test_get_lock_concurrency_locking_two_lock(
             futures.ThreadPoolExecutor, False)
@@ -906,18 +996,6 @@ class TestAPI(tests.TestWithCoordinator):
         self.assertTrue(lock.release())
 
 
-class ZakeTestAPI(TestAPI):
-    url = "zake://"
-
-
-class IPCTestAPI(TestAPI):
-    url = "ipc://"
-
-
-class FileTestAPI(TestAPI):
-    url = "file:///tmp"
-
-
 class TestHook(testcase.TestCase):
     def setUp(self):
         super(TestHook, self).setUp()
